import graph
import os
import sys
import getopt

from contig_graph import *

    
def parse_buildgraph(args):
    import build_graph
    optlist,args = getopt.getopt(args,'x',['kmer=','numreads=','out=','verbose'])
    numreads = 1000000 # default,
    outfile = None
    verbose = False
    for o,a in optlist:
        if o == '--kmer':
            graph.Kmer.set_k(int(a))
        elif o == '--numreads':
            numreads = int(a)
        elif o == '--out':
            outfile= a
        elif o == '--verbose':
            verbose=True
            
    infiles = args[:]

    G,GB = build_graph.build_sp(infiles,numreads,outfile,verbose)


def parse_alignkmers(args):
    import graph_util
    import build_graph
    optlist,args = getopt.getopt(args,'x',['kmer=','db=','ref=','out=','verbose','add=','dist=','seed='])

    verbose = False
    outfile = None
    ref = None
    db = None
    opt = graph.SimilarSequenceOptions()
    opt.k = 0
    opt.add = 0
    opt.dist = 0

    for o,a in optlist:
        if o == '--kmer':
            graph.Kmer.set_k(int(a))
        elif o == '--db':
            db = a
        elif o == '--ref':
            ref = a
        elif o == '--verbose':
            verbose=True
        elif o == '--add':
            opt.add = int(a)
        elif o == '--dist':
            opt.dist = int(a)
        elif o == '--seed':
            opt.k = int(a)
        elif o == '--out':
            outfile = a
    
    if verbose: print 'loading reference file'
    genes = graph_util.load_fasta(ref)
    if verbose: print 'found',len(genes),'reference sequences'

    if verbose: print 'loading graph file ...',; sys.stdout.flush()
    G = build_graph.load_binfile(db)
    if verbose: print ' done'

    seqv = graph.StringVector(list(genes.values()))
    outv = graph.KmerMatchVectorVector()

    if verbose: print 'aligning kmers to genes, this might take a while ..',; sys.stdout.flush()
    G.pickSimilar(seqv,outv,opt)
    if verbose: print ' done'

    if verbose: print 'saving results, ...'; sys.stdout.flush()
    ind = dict(enumerate(genes.keys()))
    rev = build_graph.revind(ind)
    build_graph.save_outv(rev,outv,outfile)
    if verbose: print ' done'


def parse_processgenes(args):
    import build_graph
    import graph_util
    import gene_process

    optlist,args = getopt.getopt(args,'x',['kmer=','db=','ref=','aln=','out=','cut=','verbose'])

    db = None
    genes = None
    aln = None
    outfile = None
    verbose = False

    for o,a in optlist:
        if o == '--kmer':
            graph.Kmer.set_k(int(a))
        elif o == '--db':
            db = a
        elif o == '--ref':
            ref = a
        elif o == '--aln':
            aln = a
        elif o == '--out':
            outfile = a
        elif o == '--verbose':
            verbose=True
        elif o == '--cut':
            cutoff = float(a)


    if verbose: print 'loading reference file'
    genes = graph_util.load_fasta(ref)
    if verbose: print 'found',len(genes),'reference sequences'

    if verbose: print 'loading graph file ...',; sys.stdout.flush()
    G = build_graph.load_binfile(db)
    if verbose: print ' done'

    iov = build_graph.iter_outv(aln)

    seqs = {}
    if verbose: print 'iterating over all genes'
    for g,l in iov:
        if verbose: print g
        seq = gene_process.full_proc(G,l,genes[g],cutoff)
        if seq != None:
            seqs[g] = seq

    if verbose: print 'saving results'
    of = open(outfile,'w')
    for g in seqs:
        of.write('>%s\n%s\n'%(g,seqs[g]))
    of.close()
        
    
def usage():
    print """
Usage: python assembler.py command options
where command is one of
   buildgraph
   alignkmers
   processgenes

options differ based on command:

buildgraph:
    --kmer=k        : kmer-size, should be no more than 31 or 63 based on
                      compilation options (default is 63)
    --numreads=nr   : number of reads, give an estimate for  the number
                      of reads in all files, this is so that we can estimate
                      the size of the Bloom filter for error correction, a
                      rough upper bound is fine
    --out=f         : output filename to save the de Bruijn graph
                      in a binary file format.
    --verbose       : print a lot of information 
    f1 ...          : a list of files, this can either be fasta or fastq,
                      gzip compressed or not

    
alignkmers:
    --kmer=k   : kmer-size used
    --db       : de Bruijn graph file in binary format
    --ref      : reference file in fasta format
    --seed     : seed size [default 20]
    --dist     : maximum mismatch allowed [default 4]
    --add      : maximum padding used [defaul 2]
    --out=f    : output file
    --verbose  : print a lot of information

processgenes:
    --kmer=k   : kmer-size used
    --db       : de Bruijn graph file in binary format
    --ref      : reference file in fasta format
    --aln      : alignment file from previous step
    --cut      : fraction of alignment to reference to accept
    --out      : output file for assembled sequences
    --verbose  : print out a lot of information  
"""


if __name__ == '__main__':
    a = sys.argv
    if len(sys.argv) > 1:
        if sys.argv[1] == 'buildgraph':
            parse_buildgraph(sys.argv[2:])
        elif sys.argv[1] == 'alignkmers':
            parse_alignkmers(sys.argv[2:])
        elif sys.argv[1] == 'processgenes':
            parse_processgenes(sys.argv[2:])
        else:
            usage()
    else:
        usage()
