import graph
import collections

class cov_filter:
    def __init__(self, G, cov, it):
        self.it = it
        self.cov = cov
        self.G = G

    def __iter__(self):
        return self

    def next(self):
        km = self.it.next()
        #print "got object"
        while not (self.cov[0] <= self.G.coverage(km) <= self.cov[1]):
            #print "cov too low",self.G.coverage(km),"wanted",self.cov
            km = self.it.next()
        return km


def contig_to_string(c):
    return ''.join([str(x)[0] for x in c] + [str(c[-1])[1:]])

class ContigGraph:
    def __init__(self):
        self.contigs = {}
        self.k = graph.cvar.Kmer_k


    def __getstate__(self):
        d = {'k':self.k, 'contigs': {}}
        for b,c in self.contigs.items():
            cobj = {'bases': c.bases, 'cov': c.cov, 'outlink': [x.bases for x in c.outlink], 'inlink': [x.bases for x in c.inlink]}
            d['contigs'][b] = cobj
        return d

    def __setstate__(self,d):
        self.d = d
        self.k = d['k']
        self.contigs = {}
        cont = d['contigs']
        for b,cobj in cont.items():
            self.contigs[b] = Contig(cobj['bases'][:],cobj['cov'][:])

        for b,cobj in cont.items():
            for o in cobj['outlink']:
                self.contigs[b].outlink.append(self.contigs[o])
            for i in cobj['inlink']:
                self.contigs[b].inlink.append(self.contigs[i])
        
        
    def strip_exons(self):
        modified = True
        while modified:
            modified = False
            for x in self.contigs.values():
                if len(x.outlink) == 2:
                    y,z = x.outlink
                    if len(y.inlink)+len(y.outlink)==2 and len(z.inlink)+len(z.outlink)==2 and y.outlink[0] == z.outlink[0]:
                        sy = y.bases[self.k:-self.k]
                        sz = z.bases[self.k:-self.k]
                        
                        print 'y:',sy
                        print 'z:',sz

                        if len(sy) == 0 and len(sz) > 0:
                            self.remove([y])
                            modified=True
                            break

                        if len(sz) == 0 and len(sy) > 0:
                            self.remove([z])
                            modified=True
                            break
                            
                    


    def printFastq(self):
        v = [(len(x.cov),x) for x in self.contigs.itervalues()]
        v.sort(reverse=True)
        
        for i,(y,x) in enumerate(v):
            print ">seq%d_%d\n%s"%(i+1,sum(x.cov)/len(x.cov), x.bases)

    def addContigs(self,G,cs, cov,debug=False):
        self.contigs = {}
        f_ends,b_ends = {},{}


        if debug:
            assert(all(all(str(c[i])[1:] == str(c[i+1])[:-1] for i in xrange(len(c)-1)) for c in cs))

        if debug:
            print "preparing contigs"

        for c in cs:
            c_cov = [G.coverage(x) for x in c]
            contig = Contig(c,c_cov)
            self.contigs[contig.bases] = contig
            if debug and str(c[0]) in f_ends:
                print "contig collision\nprevious",f_ends[str(c[0])].bases
                print "current",contig
                assert(False)
            f_ends[str(c[0])] = contig
            b_ends[str(c[-1])] = contig

        

        for c in cs:
            contig = f_ends[str(c[0])]
            for n in cov_filter(G,cov,G.fwBegin(c[-1])):
                if str(n) in f_ends:
                    if debug and f_ends[str(n)].bases[:38] != contig.bases[-38:]:
                        print "contig:", contig
                        print "end string:", str(c[-1])
                        print "n:",n
                        print "ocontig:",f_ends[str(n)]
                        assert(False)
                        
                    contig.addOutLink(f_ends[str(n)])
                    f_ends[str(n)].addInLink(contig)

            for n in cov_filter(G,cov,G.bwBegin(c[0])):
                if str(n) in b_ends:
                    contig.addInLink(b_ends[str(n)])
                    b_ends[str(n)].addOutLink(contig)


    def printDot(self):
    
        print "digraph G{\n graph [rankdir=LR];\n node =[shape=record]"
        max_cov = max(sum(c.cov)/float(len(c.cov)) for s,c in self.contigs.items())
        for s,c in self.contigs.items():
            ss = c.bases
            if len(c.bases) >= 78:
                ss = "%s .. (%d) .. %s"%(c.bases[:39],int(sum(c.cov)/float(len(c.cov))),c.bases[-39:])
            print '%s [style=filled, fillcolor=gray%s,label="%s"];'%(c.bases,int(100.0-round(70.0*(sum(c.cov)/float(len(c.cov)))/max_cov)),ss)
	
        for s,c in self.contigs.items():

            for v in c.outlink:
                print "%s -> %s;"%(c.bases,v.bases)

        print "}"

    def check(self):
        for b,c in self.contigs.items():
            assert(b==c.bases)

            for o in c.outlink:
                if not c.bases[-38:] == o.bases[:38]:
                    print "bases do not match"
                    print "c:",c
                    print "o:",o
                    print c.bases[-38:]
                    print o.bases[:38]
                    assert(False)
                if not c in o.inlink:
                    print "c not out-neighbour of o"
                    assert(False)
            for i in c.inlink:
                if not c.bases[:38] == i.bases[-38:]:
                    print "bases do not match"
                    print "c:",c
                    print "i:",i
                    print c.bases[:38]
                    print i.bases[-38:]
                    assert(False)
                if not c in i.outlink:
                    print "c not in-neighbour of o"
                    pdb.set_trace()
                    assert(False)
                
    def copy(self):
        CG = ContigGraph()
        for b,c in self.contigs.iteritems():
            CG.contigs[b] = Contig(c.bases[:] ,c.cov[:])

        for b,c in self.contigs.iteritems():
            nc = CG.contigs[b]
            for o in c.outlink:
                nc.outlink.append(CG.contigs[o.bases])
            for i in c.inlink:
                nc.inlink.append(CG.contigs[i.bases])

        return CG

    def clean(self):
        mcov = 0.0
        for c in self.contigs.itervalues():
            ccov = sum(c.cov)/float(len(c.cov))
            if ccov > mcov:
                mcov = ccov

        t = []
        for c in self.contigs.itervalues():
            if len(c.outlink) + len(c.inlink) <= 1 and sum(c.cov)/float(len(c.cov)) <= mcov/8.0 and len(c.bases) <= self.k+10:
                t.append(c)
            elif len(c.outlink)+len(c.inlink) == 0 and sum(c.cov)/float(len(c.cov)) <= mcov/4.0 and len(c.bases) <= self.k+10:
                t.append(c)

        self.remove(t)
            
    def stitch(self,G,d):
        for j in xrange(self.k-2,d-1,-1):
            #print "testing",j
            self.check()
            modified = True
            while modified:
                modified = False
                ends = {}
                for c in self.contigs.itervalues():
                    if len(c.inlink) == 0:
                        ends[c.bases[:j]] = c

                #for x in ends:
                #    print x
                #print '--'
                
                for c in self.contigs.values():
                    #print c.bases[-j:]
                    if len(c.outlink) == 0 and c.bases[-j:] in ends:
                        o = ends[c.bases[-j:]]
                        if c==o:
                            continue



                        nb = c.bases + o.bases[j:]
                        nc = Contig(nb)

                        bridge_s = [nb[i:(i+self.k)] for i in xrange(len(c.bases)-self.k+1,len(c.bases)-j)] 

                        for x in bridge_s:
                            if not all(t in 'ACGT' for t in x):
                                print x
                                assert(False)
                            if len(x) != self.k:
                                print x
                                assert(False)
                        
                        nc.cov = c.cov[:] + [G.coverage(graph.Kmer(x)) for x in bridge_s] + o.cov[:]
                        for on in o.outlink:
                            nc.outlink.append(on)
                            on.inlink.remove(o)
                            on.inlink.append(nc)

                        for ii in c.inlink:
                            nc.inlink.append(ii)
                            ii.outlink.remove(c)
                            ii.outlink.append(nc)

                        del self.contigs[c.bases]
                        del self.contigs[o.bases]
                        self.contigs[nc.bases] = nc


                        modified = True
                        break
                            

                        
            


        

    def remove(self,it):
        for x in it:
            if x.bases in self.contigs:
                #print "deleting item"
                c = self.contigs[x.bases]
                for v in c.outlink:
                    v.inlink.remove(c)
                for v in c.inlink:
                    v.outlink.remove(c)
                del self.contigs[x.bases]
            else:
                print "strangest thing happened"

    def join(self):
        modified = True
        while modified:
            modified = False
            for s,c in self.contigs.items():
                if len(c.outlink) == 1 and len(c.outlink[0].inlink) == 1:
                    o = c.outlink[0]
                    if c==o:
                        #loop remove the links and continue
                        c.outlink.remove(o)
                        c.inlink.remove(o)
                        continue
                    os = o.bases[:]
                    cs = c.bases[:]
                    nc = Contig()
                    nc.bases = c.bases + o.bases[(self.k-1):]
                    nc.cov = c.cov + o.cov
                    nc.outlink = o.outlink
                    nc.inlink = c.inlink
                    for x in nc.outlink:
                        x.inlink.remove(o)
                        x.inlink.append(nc)
                    for x in nc.inlink:
                        x.outlink.remove(c)
                        x.outlink.append(nc)

                    for no in nc.outlink:
                        if no.bases[:38] != nc.bases[-38:]:
                            pass

                    for ni in nc.inlink:
                        if ni.bases[-38:] != nc.bases[:38]:
                            pass
                    

                    del self.contigs[c.bases]
                    if not o.bases in self.contigs:
                        print len(o.bases)
                        #print self.contigs.keys()
                    else:
                        del self.contigs[o.bases]
                    self.contigs[nc.bases] = nc

                    modified = True
                    break
                    

    def removeTips(self, align=False, cov_ratio = 1.0, gaplimit = 100, max_len = 100):
        modified = True
        while modified:
            modified = False
            nodes = set(y for x,y in self.contigs.items() if len(y.outlink)+len(y.inlink) == 1)
            for tip in nodes:
                if self.removeTipsNode(tip, align,cov_ratio, gaplimit, max_len):
                    modified = True
                    break

    def removeTipsNode(self, tip, align, cov_ratio, gaplimit, max_len):
        fw = True
        if len(tip.inlink) == 1:
            fw = False

        if fw:
            o = tip.outlink[0]
            cand = [x for x in o.inlink if x != tip]
        else:
            o = tip.inlink[0]
            cand = [x for x in o.outlink if x != tip]

        remove = False
        for x in cand:
            if align:
                pass

            else:
                tipcov = sum(tip.cov)/float(len(tip.cov))
                xcov = sum(x.cov)/float(len(x.cov))

                if xcov/tipcov > cov_ratio and len(tip.bases) <= max_len:
                    remove = True

        if remove:
            self.remove((tip,))
            return True
        else:
            return False
            

    def __len__(self):
        return len(self.contigs)

    def __iter__(self):
        return self.contigs.itervalues()

    def getKmers(self):
        outv = graph.KmerVector()
        for c in self.contigs.itervalues():
            for i in xrange(len(c.bases)-self.k):
                s = c.bases[i:(i+self.k)]
                assert(len(s)==self.k)
                assert(all(x in 'ACGT' for x in s))
                outv.append(graph.Kmer(s))
        return outv
        
                

    def tourBus(self, gaplimit = 3,verbose=False):
        modified = True
        while modified:
            modified = False
            nodes = set(y for x,y in self.contigs.items() if len(y.outlink)>1)
            for y in nodes:
                if self.tourBusNode(y,gaplimit,verbose):
                    modified = True
                    break

    def tourBusNode(self,start,gaplimit,verbose):
        P = {}
        D = collections.deque([start])

        while len(D) > 0:
            p = D.popleft()

            for v in p.outlink:
                if v in P:
                    #we have a cycle
                    a = [v,p]
                    b = [v]
                    # find cycles
                    for x in [a,b]:
                        end = x[-1]
                        while end != start:
                            end = P[end]
                            x.append(end)
                        x.reverse()

                    common = -1
                    for i,(x,y) in enumerate(zip(a,b)):
                        if x != y:
                            common = i
                            break
                        
                    if common > 1:
                        continue
                    
                    if verbose:
                        print "started from",start.bases
                        print "first",common," contigs in common"

                    

                    a = a[1:-1]
                    b = b[1:-1]

                    if len(a) == 0 or len(b)==0:
                        continue

                    sa = a[0].bases[:(self.k-1)]+''.join([x.bases[(self.k-1):] for x in a])
                    sb = b[0].bases[:(self.k-1)]+''.join([x.bases[(self.k-1):] for x in b])

                    if abs(len(sa)-len(sb)) <= gaplimit:
                        if verbose: print "comparing\n%s\n%s"%(sa,sb)
                        if string_dist(sa,sb) <= gaplimit:
                            if verbose:
                                print "found cycle"
                                print "first (%d contigs)"%(len(a),), sa
                                print [x.bases for x in a]                            
                                print "second (%d contigs)"%(len(b,)),sb
                                print [x.bases for x in a]
                    
                            ## we'll remove the contigs
                            cov_a = sum(sum(x.cov) for x in a)
                            cov_b = sum(sum(x.cov) for x in b)

                            if cov_a/len(sa) > cov_b/len(sb):
                                self.remove(b)
                            else:
                                self.remove(a)


                            return True
                                
                else:
                    P[v] = p
                    D.append(v)
        return False
            
            

        

    
class Contig:

    def __init__(self, kv = None, cov = []):
        if kv == None:
            self.bases = ""
        elif isinstance(kv,str):
            self.bases = kv
        else:
            self.bases = contig_to_string(kv)
        self.cov = cov
        self.outlink = []
        self.inlink = []

    def __str__(self):
        return self.bases + " coverage: %0.2f, outDegree: %d, indegree: %d "%(sum(x for x in self.cov)/float(len(self.cov)), len(self.outlink), len(self.inlink))

    def addInLink(self,cont):
        if cont not in self.inlink:
            self.inlink.append(cont)

    def addOutLink(self,cont):
        if cont not in self.outlink:
            self.outlink.append(cont)

    
    



def contigs_to_dot(G,contigs):
    ends = {}
    cov = {}
    for c in contigs:
        ends[str(c[0])] = c
        ends[str(c[-1])] = c
        cov[str(c[0])] = sum(G.coverage(x) for x in c)/float(len(c))
    
        
    print "digraph G{\n graph [rankdir=LR];\n node =[shape=record]"
    max_cov = max(cov.values())
    for c in contigs:
    	print "%s [style=filled, fillcolor=gray%s];"%(contig_to_string(c),int(100.0-round(100.0*cov[str(c[0])]/max_cov)))
	
    for c in contigs:
        cs = contig_to_string(c)
        for n in G.fwBegin(c[-1]):
            if str(n) in ends:
                print "%s -> %s;"%(cs,contig_to_string(ends[str(n)]))
            else:
                print "%s -> %s;"%(cs,str(n))

        for n in G.bwBegin(c[0]):
            if str(n) in ends:
                pass
            else:
                print "%s -> %s;"%(str(n),cs)

    print "}"


def string_dist(a,b):
    max_len = max(len(a),len(b))
    mat = [[0]*(len(b)+1) for i in xrange(len(a)+1)]
    for i in xrange(len(a)+1):
        mat[i][0] = i
    for j in xrange(len(b)+1):
        mat[0][j] = j

    for i in xrange(1,len(a)+1):
        for j in xrange(1,len(b)+1):
            r = 1
            if a[i-1] == b[j-1]:
                r = 0
            mat[i][j] = min(mat[i-1][j-1]+r,mat[i-1][j]+1,mat[i][j-1]+1)

    return mat[len(a)][len(b)]
