%module graph
#pragma SWIG nowarn=315
%{
#include "kmer.hpp"
#include "graph.hpp"
#include "fastq.hpp"
%}

%include "std_string.i"
%include "std_vector.i"
%include "std_pair.i"

using namespace std;


%extend KmerMatch {
  %ignore KmerMatch::refpos;
  size_t refpos() const {
    return (size_t) $self->refpos;
  }
};

namespace std {
  %template(StringVector) vector<string>;
  %template(KmerVector) vector<Kmer>;
  %template(KmerMatchVector) vector<KmerMatch>;
  %template(KmerMatchVectorVector) vector<vector<KmerMatch> >;
};


%include "kmer.hpp"
%ignore int2bin;

%extend Kmer {
  char *__str__() {
    static char tmp[Kmer::MAX_K+1];
    $self->toString(tmp);
    return &tmp[0];
  }
};


%extend EdgeIterator {
  %ignore operator->();

  Kmer* _deref() {
    if ($self->pos_valid()) {
      return &(**(self));
    } else {
      return NULL;
    }
  }

  void _incr() {
    ++(*(self));
  }
  

%insert("python") %{
    def __iter__(self):
        return self

    def next(self):
        obj = self._deref()
        if obj is not None:
            ret = Kmer(obj)
            self._incr()
            return ret
        else:
            raise StopIteration
%}
};

%include "graph.hpp"
%ignore google::sparse_hash_map;
%ignore google::sparse_hash_set;
%ignore Graph::begin;
%ignore Graph::end;



namespace std {
  %template(nonConstGraphPair) pair<Kmer, graph_data>;
  %template(GraphPair) pair<const Kmer, graph_data>;
}

%include "fastq.hpp"


%inline %{
void _hmap_iter_incr( hmap_t::iterator* iter) {
  ++(*iter);
}
 %}

%extend graph_data {
  size_t coverage() {
    return (size_t) $self->cov;
  }
}

%extend Graph {

%insert("python") %{
  def __iter__(self):
    return GenericIterator(self._begin_, self._deref_key, _hmap_iter_incr)

  def items(self):
    return GenericIterator(self._begin_, self.wrap_deref_pair, _hmap_iter_incr)

  def wrap_deref_pair(self, it):
    obj = self._deref_pair(it)
    if obj is not None:
      return (obj.first,obj.second)
    else:
      return None 
%}


hmap_t::iterator* _begin_() {
    return new hmap_t::iterator(($self)->begin());
}

 const Kmer* _deref_key(const hmap_t::iterator* iter) {
   if (*iter == ($self)->end()) {
     return NULL;
   }
   
   return &((*iter)->first);
 }

 pair<const Kmer, graph_data>* _deref_pair(const hmap_t::iterator* iter) {
   if (*iter == ($self)->end()) {
     return NULL;
   }
   
   return &(**iter);
 }
};

%insert("python") %{
class GenericIterator:
    def __init__(self, begin_iter_method, deref_method, incr_method):
        self.it = begin_iter_method()
        self.incr   = incr_method
        self.deref  = deref_method

    def __iter__(self):
        return self

    def next(self):
        obj = self.deref( self.it )
        if obj is not None:
            self.incr( self.it )
            return obj
        else:
            raise StopIteration
%}
