import collections
import graph
import zlib

from contig_graph import *

def coverage_list(G,kv):
    assert(all(all(t in 'ACGT' for t in x[0]) and len(x[0]) == graph.cvar.Kmer_k) for x in kv)
    l = [(x[1],G.coverage(graph.Kmer(x[0]))) for x in kv]

    d = collections.defaultdict(int)
    for p,c in l:
        if c > d[p]:
            d[p] = c

    lp = [(p,d[p]) for p in d]

    lp.sort()
    return lp


def cov_estimate(G,gene,gene_len =1000):
    if len(gene) == 0 or len(gene[0]) != 2:
        raise ValueError()
    lp = coverage_list(G,gene)

    ll = [x[1] for x in lp]
    ll.sort()

    if len(ll)>0:
        return ll[int(len(ll)*.9)]
    else:
        return 0


def gene_lens(genes,nm_fasta):
    return dict((g,len(nm_fasta[genes[g][0][0]])) for g in genes if genes[g][0][0] in nm_fasta)

def bridge_contigs(G,contigs,cov,depth=300,verbose=False):
    global stopFlag

    try:
        stopFlag
    except NameError:
        stopFlag = False
        


    # get starting points
    fw,bw = [],[]
    targets = set()
    for c in contigs:
        fw.append(c[-1])
        bw.append(c[0])
        for x in c:
            targets.add(str(x))


    iterm = {'fw': lambda x: cov_filter(G,cov,G.fwBegin(x)), 'bw': lambda x: cov_filter(G,cov,G.bwBegin(x))}
    initm = {'fw':fw,'bw':bw}

    paths = []
    
    for d in ['fw']:
        it = iterm[d]
        initials = initm[d]

        for v in initials:
            if stopFlag:
                break
            if verbose: print "starting from ",str(v)
            stack = [(v,it(v))]
            
            while stack:
                if stopFlag:
                    break
                parent,children = stack[-1]
                try:
                    child = children.next()
                    if verbose: print ' '*len(stack)+str(child)
                    if str(child) in targets:
                        if verbose: print "found match",str(child)
                        p = [x[0] for x in stack] + [child]
                        paths.append(p)
                    else:
                        if len(stack) < depth:
                            stack.append((child,it(child)))
                except StopIteration:
                    stack.pop()

    return paths


    

        
def process_gene(G,gene, gene_len = 1000,c=None,debug=False,cc=True,depth=300,nosimp = False,verbose=False):
    if len(gene) == 0:
        if verbose:
            print "empty gene"
        return None
    gs = [x[0] for x in gene]
    

    check = all(all(x in 'ACGT' for x in t) and len(t)==graph.cvar.Kmer_k for t in gs)
    if not check:
        print "error, string not in correct format"

        for t in gene:
            if len(t) != graph.cvar.Kmer_k:
                print t

        return None

    
    if c == None:
        c = cov_estimate(G,gene)


        if verbose:
            print "c:",c
        
        
        if c >= 100:
            c = int(c/10.0),1000000
        elif c >= 50:
            c = int(max(c/10.0,5.0)),1000000
        else:
            c = 3,3000

        
    
    kv = [graph.Kmer(x) for x in gs if c[0] <= G.coverage(graph.Kmer(x)) <= c[1]]
    if verbose:
        print "coverage cutoff:",c
        print "found",len(kv),"kmers"

    comp = [(len(zlib.compress(str(x),1)),x) for x in kv]
    comp.sort()
    if len(comp)>0 and comp[0][0] < 30:
        kv = [x[1] for x in comp if x[0] >= 30]
        if verbose:
            print "trimmed down to ",len(kv),"kmers"

    if len(kv) ==0:
        return ContigGraph()

    cv = kv
    if cc:
        maxlen = max(10*gene_len,10*len(kv))
        cv = connected_component(G,kv,c, maxlen)
        if len(cv) >= maxlen:
            if verbose:
                print "Found too many kmers",len(cv)
            if not nosimp:
                return None

    if verbose:
        print "found",len(cv),"kmers"
    #old_km = set(str(x) for x in kv)
    #new_km = set(str(x) for x in cv if str(x) not in old_km)


    #old_cov = sum(G.coverage(x) for x in kv)/float(len(kv))
    #new_cov = sum(G.coverage(x) for x in cv)/float(len(kv))

    #if new_cov >= 3.0*old_cov or len(new_km) >= *len(old_km):
    #    cv = kv


    #comp = [(len(zlib.compress(str(x),1)),x) for x in cv]
    #comp.sort()
    #if comp[0][0] < 30:
    #    cv = [x[1] for x in comp if x[0] >= 30]
    
    #print "found",len(cv),"kmers using connected components"
    contigs = get_all_contigs(G,cv,c)

    p = bridge_contigs(G,contigs,c,depth=depth)
    cv = flatten(p)+cv

    contigs = get_all_contigs(G,cv,c)
    if verbose:
        print "starting with",len(contigs)

    CG = ContigGraph()
    CG.addContigs(G,contigs,c,debug)

    if nosimp == True:
        return CG
    
    if debug: CG.check()
    if verbose: print "1 size:",len(CG); 
    CG.join()
    if debug: CG.check()
    if verbose: print "2 size:",len(CG); 
    CG.removeTips()
    if debug: CG.check()
    if verbose: print "3 size:",len(CG); 
    CG.join()
    if debug: CG.check()
    if verbose: print "4 size:",len(CG); 
    CG.tourBus(gaplimit=10)
    if debug: CG.check()
    if verbose: print "5 size:",len(CG); 
    CG.join()
    if debug: CG.check()
    if verbose: print "6 size:",len(CG); 
    CG.removeTips(cov_ratio=1.5, max_len=200)
    if debug: CG.check()
    if verbose: print "7 size:",len(CG); 
    CG.removeTips(cov_ratio=2.0, max_len=500)
    if debug: CG.check()
    if verbose: print "8 size:",len(CG); 
    CG.join()
    if debug: CG.check()
    if verbose: print "9 size:",len(CG); 

    return CG

def save_setup(s):
    import cPickle
    f = open('genes/%s.pkl'%(s,),'w')
    print 'saving CGS'
    cPickle.dump(CGS,f)
    #cPickle.dump(goodl,f)
    #cPickle.dump(bad,f)
    #cPickle.dump(glen,f)
    #cPickle.dump(cov_est,f)
    f.close()

def load_setup(s):
    global CGS
    import cPickle
    f = open('genes/%s.pkl'%(s,))
    CGS = cPickle.load(f)
    #goodl = cPickle.load(f)
    #bad = cPickle.load(f)
    #glen = cPickle.load(f)
    #cov_est = cPickle.load(f)
    f.close()

def clean_and_eval(s):
    global G,CGS, res,c
    c = {}

    print "working on species ",s
    get_gseq()
    print "loading binary file"
    G = load_binfile('binfiles/%s_after.bin'%(s,))
    print "loading results"
    load_setup(s)

    print "cleaning up graphs and stitching"
    res = clean_graphs()

    sys.stdout.flush()
    print "evaluating paths"
    for g in res:
        c[g] = evalpaths(g,res[g],gseq[g])



    print "saving results"
    import cPickle
    f = open('genes/%s_res_and_paths.pkl'%(s,),'w')
    cPickle.dump(res,f)
    cPickle.dump(c,f)
    f.close()

def load_res_and_paths(s):
    global res,c
    import cPickle
    f = open('genes/%s_res_and_paths.pkl'%(s,))
    res = cPickle.load(f)
    c = cPickle.load(f)
    f.close()
    
    

def allpaths(CG):
    
    k = graph.cvar.Kmer_k
    
    # get starting points
    fw = []
    iso = []

    for c in CG.contigs.values():
        if len(c.inlink) == 0:
            if len(c.outlink) > 0:
                fw.append(c)
            else:
                iso.append(c)

    paths = []

    # for all starting points
    #print len(fw)
    for v in fw:
        stack = [(v,iter(v.outlink))]

        #print 'fw:',v
        # while we have something left
        while stack:
            parent,children = stack[-1]
            try:
                child = children.next()
                #print 'child:',len(stack),child

                # if we have children, traverse
                if len(child.outlink) > 0:
                    if child in (x for x,i in stack):
                        print >> sys.stderr, 'cycle!'
                    else:
                        stack.append((child,iter(child.outlink)))                    
                else:
                    # endpoint, add string to candidate
                    paths.append(''.join([x.bases[:-k] for x,i in stack] + [stack[-1][0].bases[-k:]]))
            except StopIteration:
                stack.pop()

    for v in iso:
        paths.append(v.bases)
    return paths

def evalpaths(g,CG,gs,verbose=False):
    import fsa
    paths = allpaths(CG)


    paths.sort(key = lambda x: len(x),reverse=True)

    if len(paths) > 50:
        paths = paths[:50]

    ret = []
    c = [0]*len(paths)
    if verbose:
        print >> sys.stderr, g,len(paths)
    for i,p in enumerate(paths):
        if len(p) > len(gs)/2.0:
            al = fsa.align(gs,[p])
            for j in xrange(len(al[0])):
                if al[0][j] != '-' and al[1][0][j] != '-':
                    c[i]+= 1
        
        c[i] = c[i]/float(len(gs))
        if c[i] > 0:
            ret.append((p,c[i]))

    return ret
        

def clean_graphs():
    global CGS,G
    res = {}

    for g in CGS:
        try:
            CG = CGS[g].copy()
            CG.stitch(G,10)
            CG.clean()
            CG.join()

            res[g] = CG
        except AssertionError, e:
            print e
    return res


def select_sequences(c,cutoff):
    ss = {}
    for g in c:
        cand = [t for t in c[g] if t[1] > cutoff]
        if len(cand) == 0:
            pass
        elif len(cand) == 1:
            ss[g] = cand[0][0]
        else:
            cand.sort(key= lambda t: t[1],reverse=True)
            ss[g] = cand[0][0]
            #if abs(cand[0][1] - cand[1][1]) < 0.1:
            #    print g


    f = open('/home/pmelsted/primSeq/penultimateResults/%s_genes.txt'%(s,),'w')
    for g in ss:
        f.write('>%s\n%s\n'%(g,ss[g]))
    f.close()
    
    return ss

        


def squish_seq(G,cand):
    if len(cand) < 2:
        return ''

    n = len(cand[0])
    l = ['-']*n
    for c in cand:
        for i,x in enumerate(c):
            if x != '-':
                if l[i] == '-':
                    l[i] = x
                else:
                    b = l[i]
                    if isinstance(b,list):
                        if x not in b:
                            b.append(x)
                    else:
                        if x != b:
                            l[i] = [b,x]
    
    amb = [i for i in xrange(n) if isinstance(l[i],list)]
    s = None
    if len(amb)==1:
        c = amb[0]
        s = [None]*len(l[c])

        for i,b in enumerate(l[c]):
            s[i] = ''.join(x for x in l[:c]+[l[c][i]]+l[c+1:] if x != '-')

    return l,amb,s

def runproc(G,l):
    stopFlag = False
    signal.alarm(30)
    x = None
    try:
        x = process_gene(G,l)
    except TimeoutException, msg:
        x = None
        print "timed out"
    finally:
        # turn off alarm
        stopFlag = False
        signal.alarm(0)
    return x

class TimeoutException(Exception): pass

def sig_handler(signum, frame):
    stopFlag = True
    raise TimeoutException("timed out")

def proc():
    global CGS,g,l,iov,G

    signal.signal(signal.SIGALRM, sig_handler)
    for g,l in iov:
        print g
        x = None
        x = runproc(G,l)

        if x != None:
            CGS[g] = x
            

def all_CGS(G,gkm,genes,glen):
    CGs = dict()
    for g in genes:
        print "processing gene",g
        gl = 1000
        if g in glen:
            gl = glen[g]
        CGs[g] = process_gene(G,gkm[g], gl)

    return CGs



def proc_all_genes(s):
    global G,CGS, iov,g,l
    g = None
    l = []
    stopFlag = False
    G = load_binfile('binfiles/%s_after.bin'%(s,))
    CGS = dict()
    iov = iter_outv(s)
    proc()
    save_setup(s)



def full_proc(G,l,g,cutoff,verbose=False):
    if len(l) == 0:
        return None
    CG = process_gene(G,l,verbose=verbose)

    if CG == None:
        return None
    
    p = evalpaths(g,CG,gseq[g],verbose=verbose)

    cand = [t for t in p if t[1] > cutoff]
    if len(cand) == 0:
        return None

    cand.sort(key=lambda t: t[1],reverse=True)
    return cand[0][0]

            
                
